#pragma rtGlobals=1		// Use modern global access method.
#pragma version=1		//November 12, 2007
#pragma IgorVersion=6

//This procedure file provides a structure and wrapper functions to add spectral estimate 
//controls to a separately programmed control panel.

//For an example of how to use the wrapper functions, see the Time-Frequency procedure file.

//This procedure file provides a windowed periodogram, Blackman-Tukey, and Welch spectral estimates.

//written by Ben Cramer (bscramer@uoregon.edu)


strconstant K_FFTWindowList="Bartlett;Blackman;cos2;Hamming;Hanning;KaiserBessel;Poisson2;Riemann"
constant K_specmaxparams=3
//References:

//Gardner, W.A., Statistical Spectral Analysis: A Nonprobabilistic Theory, 
//Prentice Hall, Englewood Cliffs, NJ, 1988, 566 pp.

//Bloomfield, P., Fourier Analysis of Time Series:  An Introduction,
//John Wiley & Sons, New York, 1976, 258 pp.

Structure SpectrumParameters
	//Universal structure to be used for input parameters to any function producing a spectral estimate
	
	wave dw,xw				//input waves for XY-pair data
	wave dw_int				//input wave for waveform data
	wave magnitude,phase		//output waves
	variable getParams		//set to 1 when calling to retrieve info for creation of controls
	
	variable Xmin,Xmax		//minimum and maximum x values to define a subrange of the data
	variable Fmin,Fmax		//minimum and maximum frequency of interest
	variable freqRes			//desired frequency resolution; if 0, default resolution (1/series length)
								//if <1/length, data will be padded to obtain desired resolution
	
	struct SpecSetVar SV			//structures holding set-variable parameters specific to a spectral estimate method
	
	struct SpecPopMenu PM		//structures holding popup menu parameters specific to a spectral estimate method
endStructure

Structure SpecSetVar
	//parameters to create a set variable control
	string name[K_specmaxparams]				//control name
	variable value	[K_specmaxparams]			//value
	variable low[K_specmaxparams],high[K_specmaxparams]	//low and high limits and increment for creating set variable control
	variable inc[K_specmaxparams]	
endStructure

Structure SpecPopMenu
	//parameters to create a popup menu control
	string name[K_specmaxparams]				//control name
	string value[K_specmaxparams]				//semicolon-separated list for popup menu
	string popStr[K_specmaxparams]				//value of selected item
	variable popNum[K_specmaxparams]			//number of selected item
endStructure

Function SPECTRUM_prototype(SPs)
	struct SpectrumParameters &SPs
	
	if(SPs.getParams)
		SPS.SV.name[0]="";SPS.SV.value[0]=nan;SPS.SV.low[0]=nan;SPS.SV.high[0]=nan;SPS.SV.inc[0]=nan
		SPS.SV.name[1]="";SPS.SV.value[1]=nan;SPS.SV.low[1]=nan;SPS.SV.high[1]=nan;SPS.SV.inc[1]=nan
		SPS.SV.name[2]="";SPS.SV.value[2]=nan;SPS.SV.low[2]=nan;SPS.SV.high[2]=nan;SPS.SV.inc[2]=nan
		SPS.PM.name[0]="";SPS.PM.value[0]="";SPS.PM.popStr[0]="";SPS.PM.popNum[0]=nan
		SPS.PM.name[1]="";SPS.PM.value[1]="";SPS.PM.popStr[1]="";SPS.PM.popNum[1]=nan
		SPS.PM.name[2]="";SPS.PM.value[2]="";SPS.PM.popStr[2]="";SPS.PM.popNum[2]=nan
	else
//		SPs.magnitude=nan;SPs.phase=nan
	endif	
end

Function SPECTRUM_Periodogram(SPs)
	struct SpectrumParameters &SPs

	if(SPs.getParams)
		SPS.SV.name[0]="";SPS.SV.value[0]=nan;SPS.SV.low[0]=nan;SPS.SV.high[0]=nan;SPS.SV.inc[0]=nan
		SPS.SV.name[1]="";SPS.SV.value[1]=nan;SPS.SV.low[1]=nan;SPS.SV.high[1]=nan;SPS.SV.inc[1]=nan
		SPS.SV.name[2]="";SPS.SV.value[2]=nan;SPS.SV.low[2]=nan;SPS.SV.high[2]=nan;SPS.SV.inc[2]=nan
		SPS.PM.name[0]="Taper Function"
		SPS.PM.value[0]="None;"+K_FFTWindowList
		SPS.PM.popStr[0]="Hamming"
		SPS.PM.popNum[0]=whichlistitem(SPS.PM.popStr[0],SPS.PM.value[0])+1
		SPS.PM.name[1]="";SPS.PM.value[1]="";SPS.PM.popStr[1]="";SPS.PM.popNum[1]=nan
		SPS.PM.name[2]="";SPS.PM.value[2]="";SPS.PM.popStr[2]="";SPS.PM.popNum[2]=nan
	else
		pgram(SPs.dw_int,pgramMagS=SPs.magnitude,win=SPs.PM.popStr[0],pad=1/SPs.freqRes/deltax(SPS.dw_int))
		SPs.magnitude=sqrt(SPs.magnitude)
	endif
end

Function SPECTRUM_Welch(SPs)
	struct SpectrumParameters &SPs

	if(SPs.getParams)
		SPS.SV.name[0]="Number of Segments";SPS.SV.value[0]=5;SPS.SV.low[0]=1;SPS.SV.high[0]=inf;SPS.SV.inc[0]=1
		SPS.SV.name[1]="Percent Overlap";SPS.SV.value[1]=50;SPS.SV.low[1]=0;SPS.SV.high[1]=100;SPS.SV.inc[1]=5
		SPS.SV.name[2]="";SPS.SV.value[2]=nan;SPS.SV.low[2]=nan;SPS.SV.high[2]=nan;SPS.SV.inc[2]=nan
		SPS.PM.name[0]="Taper Function"
		SPS.PM.value[0]="None;"+K_FFTWindowList
		SPS.PM.popStr[0]="Hamming"
		SPS.PM.popNum[0]=whichlistitem(SPS.PM.popStr[0],SPS.PM.value[0])+1
		SPS.PM.name[1]="";SPS.PM.value[1]="";SPS.PM.popStr[1]="";SPS.PM.popNum[1]=nan
		SPS.PM.name[2]="";SPS.PM.value[2]="";SPS.PM.popStr[2]="";SPS.PM.popNum[2]=nan
	else
		newdatafolder/o root:Packages
		newdatafolder/o root:Packages:WelchSpectrum
		make/o/d root:Packages:WelchSpectrum:tmp_dw,root:Packages:WelchSpectrum:tmp_magS
		wave tmp_dw=root:Packages:WelchSpectrum:tmp_dw,tmp_magS=root:Packages:WelchSpectrum:tmp_magS
		variable i,npnts,nsegs,seglen,segoffset,segolf
		npnts=numpnts(SPs.dw_int)
		nsegs=SPs.SV.value[0]
		segolf=SPs.SV.value[1]/100
		seglen=npnts/(nsegs-nsegs*segolf+segolf)
		segoffset=seglen*(1-segolf)
		for(i=0;i<nsegs;i+=1)
			duplicate/r=[i*segoffset,i*segoffset+seglen-1]/o SPs.dw_int,tmp_dw
			pgram(tmp_dw,pgramMagS=tmp_magS,win=SPs.PM.popStr[0],pad=1/SPs.freqRes/deltax(tmp_dw))
			if(i==0)
				duplicate/o tmp_magS,SPs.Magnitude
			else
				SPs.magnitude+=tmp_magS
			endif
		endfor
		SPs.magnitude=sqrt(SPs.magnitude/nsegs^2)
		copyscales/p tmp_magS,sps.magnitude
		killdatafolder/z "root:Packages:WelchSpectrum"
	endif
end

Function Pgram(dw[,pgramMagS,win,pad])
	wave dw,pgramMagS
	string win
	variable pad
		
	string PgramName
	variable np,np2,magfac
	if(paramisdefault(pgramMagS))
		PgramName=GetWavesDataFolder(dw,2)+"_pgram"
	else
		PgramName=GetWavesDataFolder(pgramMagS,2)
	endif
	make/o/d $PgramName
	wave pgramMagS=$PgramName
	duplicate/o dw,pgramMagS
	wavestats/M=1/q dw
	np=v_npnts
	pgramMagS-=v_avg
	if(ParamIsDefault(pad) || (pad==0) || numtype(pad) || np>pad)
		np2=np+mod(np,2)
	else
		pad=round(pad)
		np2=pad+mod(pad,2)
//		np2=2^(ceil(log(np2)/log(2))+1)
	endif
	if(ParamIsDefault(win) || whichlistitem(win,K_FFTWindowList)<0)
		fft/PAD={np2}/DEST=$PgramName/MAGS pgramMagS
	else
		fft/PAD={np2}/DEST=$PgramName/MAGS/WINF=$win pgramMagS
		magfac=winNorm(np,win)
		pgramMagS/=magfac
	endif
	pgramMagS/=(np/2)^2
end

function winNorm(n,windowF)
	variable n
	string windowF
	
	newdatafolder/o root:Packages
	newdatafolder/o root:Packages:winNorm
	make/o/n=(n) root:Packages:winNorm:tmpWin
	wave tmpWin=root:Packages:winNorm:tmpWin
	tmpWin=1
	windowFunction/FFT=1 $windowF,tmpWin
	wavestats/q tmpWin
	killdatafolder/Z root:Packages:winNorm
	return v_rms^2
end

Function SPECTRUM_BlackmanTukey(SPs)
	struct SpectrumParameters &SPs
	
	if(SPs.getParams)
		SPS.SV.name[0]="Lag Fraction"
		SPS.SV.value[0]=0.5
		SPS.SV.low[0]=0.05
		SPS.SV.high[0]=1
		SPS.SV.inc[0]=0.05
		SPS.SV.name[1]="";SPS.SV.value[1]=nan;SPS.SV.low[1]=nan;SPS.SV.high[1]=nan;SPS.SV.inc[1]=nan
		SPS.SV.name[2]="";SPS.SV.value[2]=nan;SPS.SV.low[2]=nan;SPS.SV.high[2]=nan;SPS.SV.inc[2]=nan
		SPS.PM.name[0]="Taper Function"
		SPS.PM.value[0]="None;"+K_FFTWindowList
		SPS.PM.popStr[0]="Hamming"
		SPS.PM.popNum[0]=whichlistitem(SPS.PM.popStr[0],SPS.PM.value[0])+1
		SPS.PM.name[1]="";SPS.PM.value[1]="";SPS.PM.popStr[1]="";SPS.PM.popNum[1]=nan
		SPS.PM.name[2]="";SPS.PM.value[2]="";SPS.PM.popStr[2]="";SPS.PM.popNum[2]=nan
	else
		BTspec(SPs.dw_int,dwMag2Name=getWavesDataFolder(SPs.magnitude,2),lagfrac=SPS.SV.value[0],winF=SPs.PM.popStr[0],pad=1/SPs.freqRes/deltax(SPs.dw_int))
		SPs.magnitude=sqrt(SPs.magnitude)
	endif	
end

Function BTspec(dw,[dw2,lagfrac,pad,winF,dwMag2Name,dwPhName,dw2Mag2name,dw2PhName,XMag2Name,XPhName,XCohName])
	wave dw,dw2;variable lagfrac,pad
	string dwMag2Name,dwPhName,dw2Mag2name,dw2PhName
	string winF,XMag2Name,XPhName,XCohName
	
	string dwFolder="",dwName="",dw2Name="",dw2Folder=""
	variable dw_avg,dw2_avg,x1,x2,dx,var,var2,cov,np,npad,lag
	
	dwFolder=GetWavesDatafolder(dw,1);dwName=nameofwave(dw)
	if(paramisdefault(dwMag2Name))
		dwMag2Name=dwFolder+dwName+"_BTMag2"
	endif
	if(paramisdefault(dwPhName))
		dwPhName=dwFolder+dwName+"_BTphase"
	endif
	if(ParamIsDefault(dw2))
		wave dw2=$getwavesdatafolder(dw,2)
	else
		dw2name=nameofwave(dw2)
		dw2Folder=getwavesdatafolder(dw2,1)
		if(ParamIsDefault(dw2Mag2Name))
			dw2Mag2Name=dw2Folder+dw2Name+"_BTMag2"
		endif
		if(ParamIsDefault(dw2PhName))
			dw2PhName=dw2Folder+dw2Name+"_BTphase"
		endif
		if(ParamIsDefault(XMag2Name))
			XMag2Name=dwFolder+dwName+"_"+dw2Name+"_XBTMag2"
		endif
		if(ParamIsDefault(XPhName))
			XPhName=dwFolder+dwName+"_"+dw2Name+"_XBTphase"
		endif
		if(ParamIsDefault(XCohName))
			XCohName=dwFolder+dwName+"_"+dw2Name+"_XBTcoh"
		endif
	endif
	NewDataFolder/o/s TMP_BTspec
	dx=max(abs(deltax(dw)),abs(deltax(dw2)))
	x1=max(pnt2x(dw,0),pnt2x(dw2,0))
	x2=min(pnt2x(dw,numpnts(dw)-1),pnt2x(dw2,numpnts(dw2)-1))
	if(x1>x2)
		doalert 0,"Waves don't overlap."
		return 0
	endif
	np=round(abs(x2-x1)/dx+1)
	if(ParamIsDefault(lagfrac))
		lag=round(np*2)
	else
		lag=ceil(lagfrac*(np*2-1))
		lag=lag-mod(lag,2)
	endif
	wavestats/q/r=(x1,x2) dw;dw_avg=v_avg;var=v_sdev^2
	wavestats/q/r=(x1,x2) dw2;dw2_avg=v_avg;var2=v_sdev^2
	cov=covariance(dw,dw2)
	make/d/o/n=(np) dwCorr,dw2Corr,XCorr
	setscale/p x x1,dx,dwCorr,dw2Corr,XCorr
	dwCorr=dw(x)-dw_avg;dw2Corr=dw2(x)-dw2_avg;XCorr=dw2(x)-dw2_avg

//calculate spectrum for dw:
	correlate dwCorr,XCorr;XCorr/=np/4
	correlate dw2Corr,dw2Corr;dw2Corr/=np/4
	correlate dwCorr,dwCorr;dwCorr/=np/4
	
	BTspec_calc(dwCorr,lag,npad=pad,win=winF);wave W_magsqr,W_phase
	duplicate/o W_magsqr,$dwMag2Name
	duplicate/o W_phase,$dwPhName
	wave dwMag2=$dwMag2Name,dwPh=$dwPhName
	//dwPh[0]=dwPh[1]//;unwrap 2*pi,dwPh
	if(strlen(dw2name))
//calculate spectrum for dw2:
		BTspec_calc(dw2Corr,lag,npad=pad,win=winF)
		duplicate/o W_magsqr,$dw2Mag2Name
		duplicate/o W_phase,$dw2PhName
		wave dw2Mag2=$dw2Mag2Name,dw2Ph=$dw2PhName
		//dw2Ph[0]=dw2Ph[1]//;unwrap 2*pi,dw2Ph
//calculate cross spectrum:
		BTspec_calc(XCorr,lag,npad=pad,win=winF)
		duplicate/o W_magsqr,$XMag2Name,$XCohName
		duplicate/o W_phase,$XPhName
		wave XMag2=$XMag2Name,XPh=$XPhName,XCoh=$XCohName
		//XPh[0]=XPh[1]//;unwrap 2*pi,XPh
		XCoh=min((XMag2)/(sqrt(dwMag2)*sqrt(dw2Mag2)),1)
		wavestats/q XMag2;v_max=(sqrt(v_max)*0.1)^2
		XPh=((XMag2>v_max)*(XCoh>0.5)) ? XPh : nan
		XCoh=((XMag2>v_max)*(XCoh>0.5)) ? XCoh : nan
	endif
	
	
	SetDataFolder ::
	killdatafolder TMP_BTspec
end

function BTspec_calc(w,lag[,npad,win])
	wave w;variable lag,npad;string win
	
	variable np,fftsign,magfac
	np=numpnts(w)
	if(ParamIsDefault(npad) || (npad==0) || numtype(npad))
		npad=np+mod(np,2)
	else
		npad=round(2*npad)
		npad=2^(ceil(ln(lag)/ln(2))+1)
	endif
	rotate -floor((np-lag)/2),w;redimension/n=(lag) w
	if(ParamIsDefault(win) || whichlistitem(win,K_FFTWindowList)<0)
		fft/PAD={npad}/DEST=fftw w
		wavetransform magsqr,fftw
		wavetransform phase,fftw
		wave W_magsqr
		W_magsqr=sqrt(W_magsqr)
	else
		fft/PAD={npad}/DEST=fftw/WINF=$win w
		wavetransform magsqr,fftw
		wavetransform phase,fftw
		wave W_magsqr
		magfac=winNorm(lag,win)
		W_magsqr=sqrt(W_magsqr)/magfac
	endif
//	fftw[1,;2]=-1*(fftw)
	
//	fftw=conj(fftw)	//??
	
//	duplicate/o fftw,fft_
	W_magsqr/=2*np
end

Function Covariance(dw1,dw2)
	wave dw1,dw2
	
	variable avg1,avg2,i,np,summ=0
	np=numpnts(dw1)
	if(numpnts(dw2)!=np)
		return nan
	endif
	wavestats/q dw1;avg1=v_avg
	wavestats/q dw2;avg2=v_avg
	for(i=0;i<np;i+=1)
		summ+=(dw1[i]-avg1)*(dw2[i]-avg2)
	endfor
	return summ/(np-1)
end

Function Multi_Taper_Method(dw,rw,ft,nt,bw,hf,fr,el,es)
	wave dw,rw,ft;variable nt,bw,hf,fr,el,es
	
	variable npts,dstep,i,j,k,l,m
	
	npts=dimsize(dw,0);dstep=dimdelta(dw,0);rw=0;ft=0
	el=trunc(el/dstep);es=trunc(es/dstep)
	make/o/n=(hf*max(el*dstep,1/fr),1) temprw=0,tempft=0;
	make/o/n=(hf*max(el*dstep,1/fr))/c tempfftsum;redimension/n=(hf/fr,1) rw,ft
	make/c/o/n=(hf*max(el*dstep,1/fr),nt) tempspec
	setscale/P x,0,fr,"cycles/"+waveunits(dw,0),rw,ft
	setscale/P x,0,min(1/(el*dstep),fr),"",temprw,tempft,tempfftsum
	make/o/d/n=(el,nt) SlepTaps=0;make/o/d/n=(nt) SlepVals=0
	Slepian_Tapers(el,nt,bw,SlepTaps,SlepVals)
	wave tapsum=tapsum;i=0;make/o tempfft;wave/c tempfftc=tempfft
	do
		duplicate/o/r=[i*es,i*es+el-1][][][] dw,tempdw
//		detrend(tempdw);j=0
		do
			m=0;tempfftsum=cmplx(0,0)
			do
				l=0
				do
					k=0
					do
						redimension/r/n=(el) tempfft;tempfft[]=tempdw[p][k][l][m]*sleptaps[p][j]
						redimension/n=(2*trunc(max(1/(2*dstep*fr),el/2))) tempfft;fft tempfft
						tempfftsum[]+=tempfft[p];k+=1
					while(k<dimsize(dw,1))
					l+=1
				while(l<dimsize(dw,2))
				m+=1
			while(m<dimsize(dw,3))
			duplicate/o tempfftsum,tempfft;tempfftc/=(k*l*m)
			tempspec[][j]=tempfftc[p]*dstep;tempfftc*=2*dstep/(SlepVals[j]*el*nt);tempfftc[0]/=2
			temprw[]+=sqrt(magsqr(tempfftc[p]))
			j+=1
		while(j<nt)
		regre2(tempspec,1+el/2,nt,tempft,tapsum,tempfft)
		redimension/n=(-1,i+1) rw,ft
		rw[][i]=sqrt(temprw[(pnt2x(rw,p)-dimoffset(temprw,0))/dimdelta(temprw,0)])
		ft[][i]=tempft[(pnt2x(rw,p)-dimoffset(tempft,0))/dimdelta(tempft,0)]
		i+=1
	while(i*es+el-1<npts)
	if(dimsize(rw,1)==1)
		redimension/n=(-1,0) rw,ft
	else
		setscale/P y,(pnt2x(dw,el/2-1)),(es*dstep),waveunits(dw,0),rw,ft
	endif
	killwaves/z tempdw,tempfft,temprw,tapsum,sleptaps,slepvals,tempspec,tempft
end

Function regre2(eigspec,nf,nt,ftest,tapsum,scratch)
	wave eigspec,ftest,tapsum;wave/c scratch;variable nf,nt

	variable sumvar,i,j,sum2
	variable/c sums

	do
		sumvar+=tapsum[i]*tapsum[i];i+=1
	while(i<nt)
	i=0
	do
		scratch[i]=cmplx(0,0);j=0
		do
			scratch[i]+=eigspec[i][j]*tapsum[j];j+=1
		while(j<nt)
		scratch[i]/=sumvar
		sum2=0;j=0
		do
			sums=eigspec[i][j]-scratch[i]*tapsum[j]
			sum2+=magsqr(sums);j+=1
		while(j<nt)
		ftest[i]=(nt-1)*magsqr(scratch[i])*sumvar/sum2
		i+=1
	while(i<nf)
end

Function Slepian_Tapers(np,nt,bw,eigvec,eigval)
	variable np,nt,bw;wave/d eigvec,eigval

//This procedure is translated more or less directly from the FORTRAN program of Mann & Lees (1996)
//which is the procedure used in the SSA-MTM toolkit (http://www.atmos.ucla.edu/tcd/ssa/)

	make/d/o/n=(np) a_1,a_2,a_3,a_4,a_5,a_6,a_7,a_8,tapsum
	make/d/i/o/n=(nt) ip
	variable ierr,dfac,drat,k,gamma,tapsq,i
	
//	ww=bw/np;cs=cos(2*pi*ww)
	a_1[]=-(cos(2*pi*bw/np))*((np-1)/2-p)^2;	a_2[]=-p*(np-p)/2;	a_3[]=a_2[p]^2

// eispack subroutines tridib and tinvit: see their documentation
	tridib(np,1e-13,a_1,a_2,a_3,0,0,1,nt,eigval,ip,ierr,a_4,a_5)
	tinvit(np,np,a_1,a_2,a_3,nt,eigval,ip,EigVec,ierr,a_4,a_5,a_6,a_7,a_8)

	dfac=pi*bw;drat=8*dfac;dfac=4*sqrt(pi*dfac)*exp(-2*dfac);k=0
	do
		eigval[k]=1-dfac
		dfac=dfac*drat/(k+1)  				//! is this correct formula? yes,but fails as k -> 2n
		k+=1
	while(k<nt)
	gamma=ln(8*np*sin(2*pi*bw/np))+0.5772156649;k=0
	do
		eigval[k]=max(1/(1+exp(pi*-2*pi*(bw-k/2-0.25)/gamma)),eigval[k])
		k+=1
	while(k<nt)
	k=0
	do
		tapsum[k]=0;tapsq=0;i=0
		do
			tapsum[k]+=eigvec[i][k];tapsq+=eigvec[i][k]*eigvec[i][k];i+=1
		while(i<np)
		tapsum[k]/=sqrt(tapsq/np);eigvec[]/=sqrt(tapsq/np);k+=1
	while(k<nt)
	killwaves/z a_1,a_2,a_3,a_4,a_5,a_6,a_7,a_8,ip
	killvariables/z ierr
end

static function tridib(n,eps1,d,e,e2,lb,ub,m11,m,w,ind,ierr,rv4,rv5)
	variable n,eps1,lb,ub,m11,m,ierr;wave/d d,e,e2,w,rv4,rv5;wave/i ind
	
	variable tag,xu,tst1,tst2,isturm,goto,m22,ii
	variable i,j,k,l,m1,m2,pp,qq,rr,ss,t1,t2,u,v,x0,x1

	ierr = 0;tag = 0;xu = d[0];x0 = d[0];u = 0;i=1
	do
		x1=u;u=0
		if(i!=n)
			u=abs(e[i])
		endif
		xu=min(d[i-1]-(x1+u),xu);x0 = max(d[i-1]+(x1+u),x0)
		if (i==1)
			e2[i-1]=0
		else
			tst1 = abs(d[i-1]) + abs(d[i-2]);tst2 = tst1 + abs(e[i-1])
			if(tst2<=tst1)
				e2[i-1]=0
			endif
		endif
		i+=1
	while(i<=n)
	x1=n;x1*=epslon(max(abs(xu),abs(x0)));xu=xu-x1;t1 = xu;x0+=x1;t2=x0;pp=1;qq=n;m1=m11-1
	do
		if(m1!=0)
			isturm = 1
			do
				v=x1;x1=xu+(x0-xu)*0.5
				if (x1== v)
					goto=980
					break
				endif
				ss=pp-1;u=1;i=pp
				do
					if (u==0)
						v = abs(e[i-1]) / epslon(1)
						if (e2[i-1] == 0)
							v = 0
						endif
					else
						v=e2[i-1]/u
					endif
					u = d[i-1] - x1 - v
					if (u<0)
						ss+=1
					endif
					i+=1
				while(i<=qq)
				if(ss==m1)
					goto=0
					break
				endif
				if(ss<m1)
					xu=x1
				else
					x0=x1
				endif
			while(1)
			if(goto==980)
				break
			endif
			xu = x1;t1 = x1
		endif
		m22=m1+m
		if(m22!=n)
			x0=t2;isturm=2
			do
				v=x1;x1=xu+(x0-xu)*0.5
				if(x1==v)
					goto=980
					break
				endif
				ss=pp-1;u=1;i=pp
				do
					if(u==0)
						v=abs(e[i-1])/epslon(1)
						if(e2[i-1]==0)
							v=0
						endif
					else
						v=e2[i-1]/u
					endif
					u=d[i-1]-x1-v
					if(u<0)
						ss+=1
					endif
					i+=1
				while(i<=qq)
				if(ss==m22)
					goto=0
					break
				endif
				if(ss<m22)
					xu=x1
				else
					x0=x1
				endif
			while(1)
			if(goto==980)
				break
			endif
			t2 = x1
		endif
		qq = 0;rr = 0
		do
			if (rr==m)
		  		goto=1001
		  		break
			endif
			tag+=1;pp=qq+1;xu=d[pp-1];x0=d[pp-1];u=0;qq=pp
			if(qq<=n)
				do
					x1=u;u=0;v=0
					if(qq!=n)
						u=abs(e[qq]);v = e2[qq]
					endif
					xu=min(d[qq-1]-(x1+u),xu);x0=max(d[qq-1]+(x1+u),x0)
					if (v==0)
						break
					endif
					qq+=1
				while(qq<=n)
			endif
			x1=epslon(max(abs(xu),abs(x0)))
			if(eps1<=0)
				eps1=-x1
			endif
			do
				if (pp==qq)
					if((t1>d[pp-1]) %| (d[pp-1]>=t2))
						break
					endif
					m1=pp;m2=pp;rv5[pp-1]=d[pp-1];ss=rr;rr+=m2-m1+1;j=1;k=m1;l=1
					do
						if(j<=ss)
							if (k>m2)
								break
							endif
							if(rv5[k-1]<w[l-1])
								ii=j
								do
									i=l+ss-ii;w[i] = w[i-1];ind[i] = ind[i-1];ii+=1
								while(ii<=ss)
								w[l-1] = rv5[k-1];ind[l-1] = tag;k+=1
							else
								j+=1
							endif
						else 
							w[l-1] = rv5[k-1];ind[l-1] = tag;k+=1
						endif
						l+=1
					while(l<=rr)
					break
				endif
				x1*=(qq-pp+1);lb=max(t1,xu-x1);ub=min(t2,x0+x1);x1=lb;isturm=3;ss=pp-1;u=1;i=pp
				do
					if(u==0)
						v=abs(e[i-1])/epslon(1)
						if(e2[i-1]==0)
							v=0
						endif
					else
						v=e2[i-1]/u
					endif
					u=d[i-1]-x1-v
					if(u<0)
						ss+=1
					endif
					i+=1
				while(i<=qq)
				m1=ss+1;x1=ub;isturm=4;ss=pp-1;u=1;i=pp
				do
					if(u==0)
						v=abs(e[i-1])/epslon(1)
						if(e2[i-1]==0)
							v=0
						endif
					else
						v=e2[i-1]/u
					endif
					u=d[i-1]-x1-v
					if(u<0)
						ss+=1
					endif
					i+=1
				while(i<=qq)
				m2=ss
				if(m1>m2)
					break
				endif
				x0=ub;isturm=5;rv5[m1-1,m2-1]=ub;rv4[m1-1,m2-1]=lb;k=m2
				do
					xu = lb;ii=m1
					do
						i=m1+k-ii
						if(xu<rv4[i-1])
							xu=rv4[i-1]
							break
						endif
						ii+=1
					while(ii<=k)
					if (x0>rv5[k-1])
						x0=rv5[k-1]
					endif
					do
						x1=(xu+x0)*0.5
						if((x0 - xu)<=abs(eps1))
							break
						endif
						tst1=2*(abs(xu)+abs(x0));tst2=tst1+(x0-xu)
						if (tst2==tst1)
							break
						endif
						ss = pp - 1;u = 1;i=pp
						do
							if(u==0)
								v = abs(e[i-1]) / epslon(1)
								if (e2[i-1]==0)
									v = 0
								endif
							else
								v = e2[i-1] / u
							endif
							u = d[i-1] - x1 - v
							if (u<0)
								ss+=1
							endif
							i+=1
						while(i<=qq)
						do
							if (ss>=k)
								x0=x1
								break
							endif
							xu = x1
							if (ss<m1)
								rv4[m1-1] = x1
								break
							endif
							rv4[ss] = x1
							if (rv5[ss-1]>x1)
								rv5[ss-1] = x1
							endif
						while(0)
					while(1)
					rv5[k-1]=x1;k-=1
					if(k<m1)
						break
					endif
				while(1)
				ss = rr;rr+=m2-m1+1;j=1;k=m1;l=1
				do
					if (j<=ss)
						if (k> m2)
							break
						endif
						if (rv5[k-1]<w[l-1])
							ii=j
							do
								i=l+s-ii;w[i]=w[i-1];ind[i]=ind[i-1];ii+=1
							while(ii<=s)
							w[l-1]=rv5[k-1];ind[l-1]=tag;k+=1
						else
							j+=1
						endif
					else 
						w[l-1] = rv5[k-1];ind[l-1] = tag;k+=1
					endif
					l+=1
				while(l<=rr)
			while(0)
			if (qq>=n)
		  		goto=1001
		  		break
		  	endif
		while(1)
	while(0)
	if(goto==980)
		ierr=3*n+isturm
	endif
	lb = t1;ub = t2
end

static function tinvit(nm,n,d,e,e2,m,w,ind,z,ierr,rv1,rv2,rv3,rv4,rv6)
	variable nm,n,m,ierr;wave/d d,e,e2,w,ind,z,rv1,rv2,rv3,rv4,rv6
	
	variable i,j,pp,qq,rr,ss,ii,ip,jj,its,tag,group,u,v,uk
	variable xu,x0,x1,eps2,eps3,eps4,norm,order,goto

	ierr = 0
	do
		if (m==0) 
			goto = 1001
			break
		endif
		tag=0;order=1-e2[0];qq = 0
		do
			pp=qq+1;qq=pp
			if(qq<=n)
				do
					if((qq==n) %| (e2[qq]==0))
						break
					endif
					qq+=1
				while(qq<=n)
			endif
			tag+=1;ss=0;rr=1
			do
				do
					if (ind[rr-1]==tag)
						its = 1;x1 = w[rr-1]
						if (ss != 0)
							if (abs(x1-x0)>=eps2)
								group=0
							else
								group+=1
								if (order * (x1 - x0)<=0)
									x1 = x0 + order * eps3
								endif
							endif
						else
							xu = 1
							if (pp==qq)
								rv6[pp-1] = 1;z[0,n-1][rr-1]=0;z[pp-1,qq-1][rr-1]=rv6[p]*xu;x0=x1
								break
							endif
							norm = abs(d[pp-1]);ip = pp +1;i=ip
							do
								norm=max(norm,abs(d[i-1])+abs(e[i-1]));i+=1
							while(i<=qq)
							eps2=1e-3*norm;eps3=epslon(norm);uk=qq-pp+1;eps4=uk*eps3;uk=eps4/sqrt(uk);ss=pp;group=0
						endif
						v=0;i=pp
						do
							rv6[i-1]=uk
							if(i!=pp)
								if (abs(e[i-1])>=abs(u))
									xu=u/e[i-1];rv4[i-1]=xu;rv1[i-2]=e[i-1];rv2[i-2]=d[i-1]-x1;rv3[i-2]=0
									if(i!= qq)
										rv3[i-2] = e[i]
									endif
									u = v-xu*rv2[i-2];v =-xu*rv3[i-2]
								else
									xu=e[i-1]/u;rv4[i-1]=xu;rv1[i-2]=u;rv2[i-2]=v;rv3[i-2]=0;u=d(i-1)-x1-xu*v
									if (i!=qq)
										v = e[i]
									endif
								endif
							else
								u = d[i-1] - x1 - xu * v
								if (i!=qq)
									v = e[i]
								endif
							endif
							i+=1
						while(i<=qq)
						if(u==0)
							u=eps3
						endif
						rv1[qq-1]=u;rv2[qq-1]=0;rv3[qq-1]=0
						do
							ii=pp
							do
								i=pp+qq-ii;rv6[i-1]=(rv6[i-1]-u*rv2[i-1]-v*rv3[i-1])/rv1[i-1];v = u;u = rv6(i-1);ii+=1
							while(ii<=qq)
							if (group!=0)
								j=rr;jj=1
								do
									do
										j-=1
										if(ind[j-1]==tag)
											break
										endif
									while(1)
									xu=0;i=pp
									do
										xu+=rv6[i-1]*z[i-1][j-1];i+=1
									while(i<=qq)
									i=pp
									do
										rv6[i-1]=rv6[i-1]-xu*z[i-1][j-1];i+=1
									while(i<=qq)
									jj+=1
								while(jj<=group)
							endif
							norm = 0;i=pp
							do
								norm+=abs(rv6[i-1]);i+=1
							while(i<=qq)
							if (norm >=1)
								goto=840
								break
							endif
							if (its==5)
								goto=830
								break
							endif
							if (norm==0)
								rv6[ss-1]=eps4;ss+=1
								if (ss>qq)
									ss = pp
								endif
							else
								xu=eps4/norm;i=pp
								do
									rv6[i-1]=rv6[i-1]*xu;i+=1
								while(i<=qq)
							endif
							i=ip
							do
								u = rv6[i-1]
								if (rv1[i-2]==e[i-1])
									u=rv6[i-2];rv6[i-2]=rv6[i-1]
								endif
								rv6[i-1]=u-rv4[i-1]*rv6[i-2];i+=1
							while(i<=qq)
							its+=1
						while(1)
						if(goto==830)
							ierr=-rr;xu=0
						else
							if(goto==840)
								u=0;i=pp
								do
									u=pythag(u,rv6[i-1]);i+=1
								while(i<=qq)
								xu=1/u
							endif
						endif
						goto=0
						z[0,n-1][rr-1]=0;z[pp-1,qq-1][rr-1]=rv6[p]*xu;x0 = x1
					endif
				while(0)
				rr+=1
			while(rr<=m)
			if(qq>=n)
		  		break
			endif
		while(1)
	while(0)
end

static function epslon (x)
	variable x
	
	variable a,b,c,eps
	
	a = 4/3
	do
		b=a-1;c = b+b+b;eps=abs(c-1)
		if(eps!=0)
			break
		endif
	while(1)
	return eps*abs(x)
end

static function pythag(a,b)
	variable a,b
	
	variable pp,rr,ss,tt,u
	
	pp = max(abs(a),abs(b))
	if (pp==0)
		return pp
	else
		rr = (min(abs(a),abs(b))/pp)^2
		do
			tt = 4 + rr
			if (tt ==4)
				break
			endif
			ss = rr/tt;u = 1 + 2*ss;pp = u*pp;rr = (ss/u)^2 * rr
		while(1)
	endif
	return pp
end

